import cherrypy
import json
import logging
import os

from cryptojwt.utils import as_bytes

from fedservice.metadata_api.db import db_make_entity_statement
from fedservice.metadata_api.fs import make_entity_statement
from fedservice.metadata_api.fs import mk_path

logger = logging.getLogger(__name__)


class MetaAPI(object):
    def __init__(self, static_dir='static'):
        self.static_dir = static_dir

    @cherrypy.expose
    def index(self):
        response = [
            '<html><head>',
            '<title>My metadata API endpoint</title>',
            '<link rel="stylesheet" type="text/css" href="/static/theme.css">'
            '</head><body>'
            "<h1>Welcome to my metadata API endpoint</h1>",
            '</body></html>'
            ]
        return '\n'.join(response)

    @cherrypy.expose
    def resolve_metadata(self, **kwargs):
        """
        Evaluate a trust path of another entity
        TODO implement

        :param kwargs:
        :return:
        """
        return "OK"

    @cherrypy.expose
    def listing(self, **kwargs):
        raise NotImplementedError()

    def entity_statement(self, **kwargs):
        raise NotImplementedError()

    @cherrypy.expose
    def metadata_api(self, **kwargs):
        try:
            _op = kwargs['op']
        except KeyError:
            _op = "fetch"

        if _op == 'fetch':
            return self.entity_statement(**kwargs)
        elif _op == 'resolve_metadata':
            return self.resolve_metadata(**kwargs)
        elif _op == 'listing':
            return self.listing(**kwargs)
        else:
            raise cherrypy.HTTPError(400, "Unknown operation")

    def _cp_dispatch(self, vpath):
        # Only get here if vpath != None
        ent = cherrypy.request.remote.ip
        logger.info('ent:{}, vpath: {}'.format(ent, vpath))

        if vpath[0] in self.static_dir:
            return self
        elif len(vpath) == 2:
            a = vpath.pop(0)
            b = vpath.pop(0)
            if a == '.well-known' and b == 'openid-federation':
                return self.metadata_api

        return self


class MetaAPIFS(MetaAPI):
    def __init__(self, base_url, data_dir='.', static_dir='static'):
        MetaAPI.__init__(self, static_dir)
        self.base_url = base_url
        self.data_dir = data_dir

    @cherrypy.expose
    def entity_statement(self, **kwargs):
        jws = make_entity_statement(self.base_url, self.data_dir, **kwargs)
        cherrypy.response.headers['Content-Type'] = 'application/json'
        return as_bytes(json.dumps(jws))

    @cherrypy.expose
    def listing(self, **kwargs):
        """
        List the subordinates of this intermediate
        TODO implement

        :param kwargs:
        :return:
        """
        try:
            iss = kwargs['iss']
        except KeyError:
            raise cherrypy.HTTPError(400, 'Missing required argument')

        _iss_dir = mk_path(self.data_dir, iss)
        if not os.path.isdir(_iss_dir):
            raise cherrypy.HTTPError(message='No such issuer')

        return "OK"


class MetaAPIDb(MetaAPI):
    def __init__(self, authn_info, db_uri, key_jar, issuer, static_dir='static',
                 lifetime=14400, sign_alg='ES256'):
        MetaAPI.__init__(self, static_dir)
        self.authn_info = authn_info
        self.db_uri = db_uri
        self.key_jar = key_jar
        self.issuer = issuer
        self.lifetime = lifetime
        self.sign_alg = sign_alg

    @cherrypy.expose
    def entity_statement(self, **kwargs):
        if 'iss' in kwargs:
            if kwargs['iss'] != self.issuer:
                raise cherrypy.HTTPError(
                    403, 'Not prepared to sign as "{}"'.format(kwargs['iss']))
        else:
            kwargs['iss'] = self.issuer

        jws = db_make_entity_statement(self.db_uri, key_jar=self.key_jar,
                                       lifetime=self.lifetime,
                                       sign_lag=self.sign_alg,
                                       authn_info=self.authn_info, **kwargs)
        cherrypy.response.headers['Content-Type'] = 'application/json'
        return as_bytes(json.dumps(jws))



class FedServiceError(Exception):
    pass


class NoSuitableFederation(FedServiceError):
    pass


class NoTrustedClaims(FedServiceError):
    pass


class DbFault(FedServiceError):
    pass


class WrongSubject(FedServiceError):
    pass


class ConstraintError(FedServiceError):
    pass


class UnknownCertificate(FedServiceError):
    pass



from cryptojwt import JWT

from fedservice import branch2lists
from fedservice import eval_chain
from fedservice.entity_statement.collect import verify_self_signed_signature
from fedservice.exception import ConstraintError
from fedservice.message import TrustMark


def create_trust_mark(entity_id, key_jar, trust_mark_id, subject='', lifetime=0, trust_mark='',
                      reference=''):
    """
    Create Trust Mark.

    :param entity_id: The issuers entity_id
    :param key_jar: A KeyJar that contains useful keys
    :param trust_mark_id: The Trust Mark identifier
    :param subject: The subject's id
    :param lifetime: For how long the trust mark should be valid (0=for ever)
    :param trust_mark: A URL pointing to a graphic trust mark
    :param reference: A URL pointing to reference material for this trust mark
    :return: A signed JWT containing the provided information
    """
    _tm = TrustMark(
        id=trust_mark_id,
    )
    if trust_mark:
        _tm["mark"] = trust_mark
    if reference:
        _tm["ref"] = reference

    if subject:
        _tm['sub'] = subject
    else:
        _tm['sub'] = entity_id

    # Create the Signed JWT representing the Trust Mark
    _jwt = JWT(key_jar=key_jar, iss=entity_id, lifetime=lifetime)
    return _jwt.pack(_tm)


def unpack_trust_mark(token, keyjar, entity_id):
    _jwt = JWT(key_jar=keyjar, msg_cls=TrustMark, allowed_sign_algs=["RS256"])
    _tm = _jwt.unpack(token)
    _tm.verify(entity_id=entity_id)
    return _tm


def get_trust_mark(federation_entity, token, entity_id, trust_anchor_id):
    _tm = unpack_trust_mark(token, federation_entity.key_jar, entity_id)

    # Get the self-signed entity statement
    entity_config = federation_entity.get_configuration_information(_tm["iss"])

    # Collect the trust chains and verify them
    statements = federation_entity.collect_metadata_statements(entity_config,
                                                               "federation_entity")
    # one of the statement chains has to end in the trust_anchor
    statement = None
    for s in statements:
        if s.iss_path[-1] == trust_anchor_id:
            statement = s
            break

    # If the trust mark is not self-signed then there should be no intermediate
    if _tm["sub"] != _tm["iss"]:
        if len(statement.iss_path) > 2:  # should only be self or self and trust anchor
            raise ConstraintError("Trust chain too long")

    return _tm





import logging

import oidcrp

logger = logging.getLogger(__name__)


class RPHandler(oidcrp.RPHandler):
    def __init__(self, base_url='', hash_seed="", keyjar=None, verify_ssl=True,
                 services=None, service_factory=None, client_configs=None,
                 client_authn_factory=None, client_cls=None,
                 state_db=None, federation_entity=None, httpc_params=None, **kwargs):
        oidcrp.RPHandler.__init__(self, base_url=base_url, hash_seed=hash_seed, keyjar=keyjar,
                                  verify_ssl=verify_ssl, services=services,
                                  service_factory=service_factory, client_configs=client_configs,
                                  client_authn_factory=client_authn_factory, client_cls=client_cls,
                                  state_db=state_db, httpc_params=httpc_params, **kwargs)

        self.federation_entity = federation_entity

    def init_client(self, issuer):
        client = oidcrp.RPHandler.init_client(self, issuer)
        client.service_context.federation_entity = self.federation_entity
        return client



def calculate_path_length(constraints, current_max_path_length, max_assigned):
    try:
        _max_len = constraints['max_path_length']
    except KeyError:
        if max_assigned:
            current_max_path_length -= 1
            return current_max_path_length

        return 0

    if max_assigned:
        new_current = current_max_path_length - 1
        if _max_len < new_current:
            return _max_len
        else:
            return new_current

    return _max_len


def remove_scheme(url):
    if url.startswith('https://'):
        return url[8:]
    elif url.startswith('http://'):
        return url[7:]
    else:
        raise ValueError('Wrong scheme: %s', url)


def more_specific(a, b):
    a_part = remove_scheme(a).split('.')
    b_part = remove_scheme(b).split('.')
    if len(a_part) >= len(b_part):
        a_part.reverse()
        b_part.reverse()
        for _x, _y in zip(a_part, b_part):
            if _x != _y:
                if _y == "":
                    return True
                return False
        return True
    return False


# def add_permitted(new_permitted, permitted):
#     _updated = []
#     for _new in new_permitted:
#         for _old in permitted:
#             if more_specific(_new, _old):
#                 _updated.append(_new)
#             else:
#                 _updated.append(_old)
#     return _updated


def update_specs(new_constraints, old_constraints):
    _updated = []
    _replaced = False
    for _old in old_constraints:
        _replaced = False
        for _new in new_constraints:
            if more_specific(_new, _old):
                _updated.append(_new)
                _replaced = True

        if not _replaced:
            _updated.append(_old)
    return _updated


def add_constraints(new_constraints, naming_constraints):
    for key in ['permitted','excluded']:
        if not naming_constraints[key]:
            if key in new_constraints and new_constraints[key]:
                naming_constraints[key] = new_constraints[key][:]

            continue
        else:
            if not new_constraints[key]:
                continue

        naming_constraints[key] = update_specs(new_constraints[key], naming_constraints[key])

    return naming_constraints


def update_naming_constraints(constraints, naming_constraints):
    try:
        new_constraints = constraints['naming_constraints']
    except KeyError:
        pass
    else:
        naming_constraints = add_constraints(new_constraints, naming_constraints)

    return naming_constraints


def excluded(subject_id, excluded_ids):
    for excl in excluded_ids:
        if more_specific(subject_id, excl):
            return True
    return False


def permitted(subject_id, permitted_id):
    for perm in permitted_id:
        if more_specific(subject_id, perm):
            return True
    return False


def meets_restrictions(trust_chain):
    """
    Verfies that the trust chain fulfills the constraints specified in it.

    :param trust_chain: A sequence of entity statements. The order is such that the leaf's is the
        last. The trust anchor's the first.
    :return: True is the constraints are fulfilled. False otherwise
    """

    current_max_path_length = 0
    max_assigned = False
    naming_constraints = {
        "permitted": None,
        "excluded": None
    }

    for statement in trust_chain[:-1]:  # All but the last
        try:
            _constraints = statement['constraints']
        except KeyError:
            continue

        current_max_path_length = calculate_path_length(_constraints, current_max_path_length,
                                                        max_assigned)

        if current_max_path_length < 0:
            return False

        naming_constraints = update_naming_constraints(_constraints, naming_constraints)

        # if explicitly excluded return False
        if 'excluded' in naming_constraints and naming_constraints['excluded']:
            if excluded(statement['sub'], naming_constraints['excluded']):
                return False

        # If there is a list of permitted it must be in there
        if 'permitted' in naming_constraints and naming_constraints['permitted']:
            if not permitted(statement['sub'], naming_constraints["permitted"]):
                return False

    # Now check the leaf entity
    statement = trust_chain[-1]
    # if explicitly excluded return False
    if 'excluded' in naming_constraints and naming_constraints['excluded']:
        if excluded(statement['sub'], naming_constraints['excluded']):
            return False

    # If there is a list of permitted it must be in there
    if 'permitted' in naming_constraints and naming_constraints['permitted']:
        if not permitted(statement['sub'], naming_constraints["permitted"]):
            return False

    return True



import logging

from cryptojwt import KeyBundle
from cryptojwt.jws.jws import factory

from fedservice.entity_statement.policy import apply_policy
from fedservice.entity_statement.policy import gather_policies
from fedservice.entity_statement.statement import Statement

logger = logging.getLogger(__name__)


def verify_trust_chain(es_list, key_jar):
    """

    :param es_list: List of entity statements. The entity's self-signed statement last.
    :param key_jar: A KeyJar instance
    :return: A sequence of verified entity statements
    """
    ves = []
    n = len(es_list) - 1
    for es in es_list:
        _jwt = factory(es)
        if _jwt:
            logger.debug("JWS header: %s", _jwt.headers())
            keys = key_jar.get_jwt_verify_keys(_jwt.jwt)
            _key_spec = ['{}:{}:{}'.format(k.kty, k.use, k.kid) for k in keys]
            logger.debug("Possible verification keys: %s", _key_spec)
            res = _jwt.verify_compact(keys=keys)
            logger.debug(("Verified entity statement: %s", res))
            try:
                _jwks = res['jwks']
            except KeyError:
                if len(ves) != n:
                    raise ValueError('Missing signing JWKS')
            else:
                _kb = KeyBundle(keys=_jwks['keys'])
                try:
                    old = key_jar.get_issuer_keys(res['sub'])
                except KeyError:
                    key_jar.add_kb(res['sub'], _kb)
                else:
                    new = [k for k in _kb if k not in old]
                    if new:
                        _key_spec = ['{}:{}:{}'.format(k.kty, k.use, k.kid) for k in new]
                        logger.debug("New keys added to the key jar: %s", _key_spec)
                        # Only add keys to the KeyJar if they are not already there.
                        _kb.set(new)
                        key_jar.add_kb(res['sub'], _kb)

            ves.append(res)

    return ves


def trust_chain_expires_at(ves):
    exp = -1
    for v in ves:
        if exp >= 0:
            if v['exp'] < exp:
                exp = v['exp']
        else:
            exp = v['exp']
    return exp


def eval_policy_chain(chain, key_jar, entity_type):
    """

    :param chain: A trust chain
    :param key_jar: A key Jar
    :param entity_type:
    :return: tuple with federation ID, combined metadata policy and expiration time
    """
    ves = verify_trust_chain(chain, key_jar)
    tp_exp = trust_chain_expires_at(ves)

    # Combine the metadata policies from the trust root and all intermediates
    return ves[0]["iss"], gather_policies(ves[:-1], entity_type), tp_exp


def eval_chain(chain, key_jar, entity_type, apply_policies=True):
    """

    :param chain: A chain of entity statements
    :param key_jar: A :py:class:`cryptojwt.key_jar.KeyJar` instance
    :param entity_type: Which type of metadata you want returned
    :param apply_policies: Apply policies to the metadata or not
    :returns: A Statement instances
    """
    logger.debug("Evaluate trust chain")
    ves = verify_trust_chain(chain, key_jar)
    tp_exp = trust_chain_expires_at(ves)

    statement = Statement(exp=tp_exp, verified_chain=ves)

    if apply_policies:
        if len(ves) > 1:
            # Combine the metadata policies from the trust root and all intermediates
            combined_policy = gather_policies(ves[:-1], entity_type)
            logger.debug("Combined policy: %s", combined_policy)
            try:
                metadata = ves[-1]['metadata'][entity_type]
            except KeyError:
                statement.metadata = None
            else:
                # apply the combined metadata policies on the metadata
                statement.metadata = apply_policy(metadata, combined_policy)
                statement.combined_policy = combined_policy
        else:
            statement.metadata = ves[0]["metadata"][entity_type]
            statement.combined_policy = {}
    else:
        # accept what ever is in the statement provided by the leaf entity
        statement.metadata = ves[-1]

    iss_path = [x['iss'] for x in ves]
    statement.fo = iss_path[0]

    iss_path.reverse()

    statement.iss_path = iss_path
    return statement



import logging

from oidcendpoint.oidc import registration
from oidcmsg.oidc import RegistrationRequest

from fedservice.entity_statement.policy import diff2policy
from fedservice.entity_statement.utils import create_authority_hints

logger = logging.getLogger(__name__)


class Registration(registration.Registration):
    request_format = 'jose'
    request_placement = 'body'
    response_format = 'jose'

    def __init__(self, endpoint_context, **kwargs):
        registration.Registration.__init__(self, endpoint_context, **kwargs)
        self.post_construct.append(self.create_entity_statement)

    def parse_request(self, request, auth=None, **kwargs):
        return request

    def process_request(self, request=None, **kwargs):
        """

        :param request: An entity statement in the form of a signed JT
        :param kwargs:
        :return:
        """
        _fe = self.endpoint_context.federation_entity

        statements = _fe.collect_metadata_statements(request, 'openid_relying_party')

        _fe.proposed_authority_hints = create_authority_hints(
            _fe.authority_hints, statements)

        statement = _fe.pick_metadata(statements)

        # handle the registration request as in the non-federation case.
        req = RegistrationRequest(**statement.metadata)
        response_info = registration.Registration.process_request(self, req, authn=None, **kwargs)
        if "response_args" in response_info:
            payload = _fe.get_payload(request)
            _policy = diff2policy(response_info['response_args'],
                                  payload['metadata'][_fe.opponent_entity_type])
            entity_statement = _fe.create_entity_statement(
                _fe.entity_id,
                payload['iss'],
                _fe.key_jar,
                metadata={'federation_entity': {"trust_anchor_id": statement.fo}},
                metadata_policy={_fe.opponent_entity_type: _policy},
                aud=payload['iss']
            )
            response_info["response_msg"] = entity_statement
            del response_info["response_args"]

        return response_info

    @staticmethod
    def create_entity_statement(response_args, request, endpoint_context,
                                **kwargs):
        """
        wrap the non-federation response in a federation response

        :param response_args:
        :param request:
        :param endpoint_context:
        :param kwargs:
        :return:
        """
        _fe = endpoint_context.federation_entity
        _md = {_fe.opponent_entity_type: response_args.to_dict()}
        return _fe.create_entity_statement(_fe.entity_id, sub=_fe.entity_id,
                                           metadata=_md,
                                           authority_hints=_fe.proposed_authority_hints)



import logging

from cryptojwt.jwt import JWT

logger = logging.getLogger(__name__)


def create_entity_statement(iss, sub, key_jar, metadata=None, metadata_policy=None,
                            authority_hints=None, lifetime=86400, aud='', include_jwks=True,
                            constraints=None, **kwargs):
    """

    :param iss: The issuer of the signed JSON Web Token
    :param sub: The subject which the metadata describes
    :param key_jar: A KeyJar instance
    :param metadata: The entity's metadata organised as a dictionary with the
        entity type as key
    :param metadata_policy: Metadata policy
    :param authority_hints: A dictionary with immediate superiors in the
        trust chains as keys and lists of identifier of trust roots as values.
    :param lifetime: The life time of the signed JWT.
    :param aud: Possible audience for the JWT
    :param include_jwks: Add JWKS
    :param constraints: A dictionary with constraints.
    :return: A signed JSON Web Token
    """

    msg = {'sub': sub}
    if metadata:
        msg['metadata'] = metadata

    if metadata_policy:
        msg['metadata_policy'] = metadata_policy

    if authority_hints:
        msg['authority_hints'] = authority_hints

    if aud:
        msg['aud'] = aud

    if constraints:
        msg['constraints'] = constraints

    if kwargs:
        msg.update(kwargs)

    if include_jwks:
        # The public signing keys of the subject
        msg['jwks'] = key_jar.export_jwks(issuer=sub)

    packer = JWT(key_jar=key_jar, iss=iss, lifetime=lifetime)

    return packer.pack(payload=msg)



import logging

from cryptojwt.jwt import utc_time_sans_frac

__author__ = 'roland'

logger = logging.getLogger(__name__)


class Statement(object):
    """
    Class in which to store the parse result from applying metadata policies on a
    metadata statement.
    """

    def __init__(self, exp=0, signing_keys=None, verified_chain=None):
        """
        :param exp: Expiration time
        """
        self.fo = ""
        self.iss_path = []
        self.err = {}
        self.metadata = {}
        self.exp = exp
        self.signing_keys = signing_keys
        self.verified_chain = verified_chain

    def keys(self):
        return self.metadata.keys()

    def items(self):
        return self.metadata.items()

    def __getitem__(self, item):
        return self.metadata[item]

    def __contains__(self, item):
        return item in self.metadata

    def claims(self):
        """
        The result after flattening the statements
        """
        return self.metadata

    def is_expired(self):
        now = utc_time_sans_frac()
        if self.exp < now:
            logger.debug('is_expired: {} < {}'.format(self.exp, now))
            return True
        else:
            return False




from oidcendpoint.oidc import registration


def add_automatic_registration_support(endpoint, **kwargs):
    """

    :param endpoint:
    :param kwargs:
    :return:
    """
    for endp in ["authorization", "pushed_authorization"]:
        auth_endpoint = endpoint.get(endp)
        if auth_endpoint:
            auto_reg = registration.Registration(auth_endpoint.endpoint_context, **kwargs)
            auth_endpoint.automatic_registration_endpoint = auto_reg

            # if auth_endpoint.client_authn_method:
            #     if "private_key_jwt" not in auth_endpoint.client_authn_method:
            #         auth_endpoint.client_authn_method.append("private_key_jwt")
            # else:
            #     auth_endpoint.client_authn_method = ['private_key_jwt']



import logging

from cryptojwt.jwk import pems_to_x5c
from oidcendpoint.oidc import provider_config
from oidcmsg import oidc
from oidcmsg.oidc import ProviderConfigurationResponse

logger = logging.getLogger(__name__)


class ProviderConfiguration(provider_config.ProviderConfiguration):
    request_cls = oidc.Message
    response_cls = ProviderConfigurationResponse
    request_format = 'jws'
    response_format = 'jws'
    endpoint_name = 'discovery'

    def __init__(self, endpoint_context, **kwargs):
        provider_config.ProviderConfiguration.__init__(self, endpoint_context,
                                                       **kwargs)
        self.post_construct.append(self.create_entity_statement)

    def process_request(self, request=None, **kwargs):
        return {'response_args': self.endpoint_context.provider_info.copy()}

    def create_entity_statement(self, request_args, request=None, **kwargs):
        """
        Create a self signed entity statement

        :param request_args:
        :param request:
        :param kwargs:
        :return:
        """

        _fe = self.endpoint_context.federation_entity
        _md = {_fe.entity_type: request_args.to_dict()}
        if _fe.collector.use_ssc:
            with open(_fe.collector.web_cert_path,'r') as fp:
                pem_cert = fp.read()
            x5c = pems_to_x5c([pem_cert])
            return _fe.create_entity_statement(_fe.entity_id, sub=_fe.entity_id, metadata=_md,
                                               x5c=x5c)
        else:
            return _fe.create_entity_statement(_fe.entity_id, sub=_fe.entity_id, metadata=_md)



import logging

from cryptojwt.jwt import utc_time_sans_frac

logger = logging.getLogger(__name__)


class ESCache(object):
    def __init__(self, allowed_delta=300):
        self._db = {}
        self.allowed_delta = allowed_delta

    def __setitem__(self, key, value):
        self._db[key] = value

    def __getitem__(self, item):
        try:
            statement = self._db[item]
        except KeyError:
            return None
        else:
            if isinstance(statement, dict):
                # verify that the statement is recent enough
                _now = utc_time_sans_frac()
                if _now < (statement["exp"] - self.allowed_delta):
                    return statement
                else:
                    del self._db[item]
                    return None
            else:
                return statement

    def __delitem__(self, key):
        del self._db[key]

    def __contains__(self, item):
        _val = self[item]
        if _val:
            return True
        else:
            return False



import logging

logger = logging.getLogger(__name__)


class PolicyError(Exception):
    pass


def combine_subset_of(s1, s2):
    return list(set(s1).intersection(set(s2)))


def combine_superset_of(s1, s2):
    return list(set(s1).intersection(set(s2)))


def combine_one_of(s1, s2):
    return list(set(s1).intersection(set(s2)))


def combine_add(s1, s2):
    if isinstance(s1, list):
        set1 = set(s1)
    else:
        set1 = {s1}
    if isinstance(s2, list):
        set2 = set(s2)
    else:
        set2 = {s2}
    return list(set1.union(set2))


POLICY_FUNCTIONS = {"subset_of", "superset_of", "one_of", "add", "value", "default", "essential"}

OP2FUNC = {
    "subset_of": combine_subset_of,
    "superset_of": combine_superset_of,
    "one_of": combine_one_of,
    "add": combine_add,
}

ALLOWED_COMBINATIONS = [
    {"superset_of", "subset_of"},
    {"default", "one_of"},
    {"default", "subset_of"},
    {"essential", "subset_of"},
    {"default", "superset_of"},
    {"essential", "superset_of"},
    {"essential", "one_of"},
    {"essential", "add"},
    {"essential", "value"},
    {"essential", "default"},
    {"default", "subset_of", "superset_of"},
    {"essential", "subset_of", "superset_of"}
]


def weed(policy):
    """
    Remove policy functions that are not part of the standard

    :param policy: Policy definition
    :return: Policy definition with all non-standard keys weeded out
    """
    return set(policy.keys()).difference(POLICY_FUNCTIONS)


def do_sub_one_super_add(superior, child, policy):
    if policy in superior and policy in child:
        comb = OP2FUNC[policy](superior[policy], child[policy])
        if comb:
            return comb
        else:
            raise PolicyError("Value sets doesn't overlap")
    elif policy in superior:
        return superior[policy]
    elif policy in child:
        return child[policy]


def do_value(superior, child, policy):
    if policy in superior and policy in child:
        if superior[policy] == child[policy]:
            return superior[policy]
        else:
            raise PolicyError("Not allowed to combine values")
    elif policy in superior:
        return superior[policy]
    elif policy in child:
        return child[policy]


def do_default(superior, child, policy):
    # A child's default can not override a superiors
    if policy in superior and policy in child:
        if superior['default'] == child['default']:
            return superior['default']
        else:
            raise PolicyError("Not allowed to change default")
    elif policy in superior:
        return superior[policy]
    elif policy in child:
        return child[policy]


def do_essential(superior, child, policy):
    # essential: an child can make it True if a superior has states False
    # but not the other way around

    if policy in superior and policy in child:
        if not superior[policy] and child['essential']:
            return True
        else:
            return superior[policy]
    elif policy in superior:
        return superior[policy]
    elif policy in child:  # Not in superior is the same as essential=True
        return True


DO_POLICY = {
    "superset_of": do_sub_one_super_add,
    "subset_of": do_sub_one_super_add,
    "one_of": do_sub_one_super_add,
    "add": do_sub_one_super_add,
    "value": do_value,
    "default": do_default,
    "essential": do_essential
}


def combine_claim_policy(superior, child):
    """
    Combine policy rules

    :param superior: Superior policy
    :param child: Intermediates policy
    """

    comb_policy = set(superior.keys()).union(set(child.keys()))
    comb_policy = comb_policy.intersection(POLICY_FUNCTIONS)
    if len(comb_policy) > 3:
        raise PolicyError("Not allowed combination of policies")
    elif len(comb_policy) in [2, 3]:
        if comb_policy not in ALLOWED_COMBINATIONS:
            raise PolicyError("Not allowed combination of policies")

    rule = {}
    for policy in comb_policy:
        rule[policy] = DO_POLICY[policy](superior, child, policy)

    if comb_policy == {'superset_of', 'subset_of'}:
        # make sure the subset_of is a superset of superset_of.
        if set(rule['superset_of']).difference(set(rule['subset_of'])):
            raise PolicyError('superset_of not a super set of subset_of')
    elif comb_policy == {'superset_of', 'subset_of', 'default'}:
        # make sure the subset_of is a superset of superset_of.
        if set(rule['superset_of']).difference(set(rule['subset_of'])):
            raise PolicyError('superset_of not a super set of subset_of')
        if set(rule['default']).difference(set(rule['subset_of'])):
            raise PolicyError('default not a sub set of subset_of')
        if set(rule['superset_of']).difference(set(rule['default'])):
            raise PolicyError('default not a super set of subset_of')
    elif comb_policy == {'subset_of', 'default'}:
        if set(rule['default']).difference(set(rule['subset_of'])):
            raise PolicyError('default not a sub set of subset_of')
    elif comb_policy == {'superset_of', 'default'}:
        if set(rule['superset_of']).difference(set(rule['default'])):
            raise PolicyError('default not a super set of subset_of')
    elif comb_policy == {'one_of', 'default'}:
        if isinstance(rule['default'], list):
            if set(rule['default']).difference(set(rule['one_of'])):
                raise PolicyError('default not a super set of one_of')
        else:
            if {rule['default']}.difference(set(rule['one_of'])):
                raise PolicyError('default not a super set of one_of')
    return rule


def combine_policy(superior, child):
    res = {}
    sup_set = set(superior.keys())
    chi_set = set(child.keys())

    for claim in set(sup_set).intersection(chi_set):
        res[claim] = combine_claim_policy(superior[claim], child[claim])

    for claim in sup_set.difference(chi_set):
        res[claim] = superior[claim]

    for claim in chi_set.difference(sup_set):
        res[claim] = child[claim]

    return res


def gather_policies(chain, entity_type):
    """
    Gather and combine all the metadata policies that are defined in the trust chain
    :param chain: A list of Entity Statements
    :return: The combined metadata policy
    """

    try:
        combined_policy = chain[0]['metadata_policy'][entity_type]
    except KeyError:
        combined_policy = {}

    for es in chain[1:]:
        try:
            child = es['metadata_policy'][entity_type]
        except KeyError:
            pass
        else:
            combined_policy = combine_policy(combined_policy, child)

    return combined_policy


def union(val1, val2):
    if isinstance(val1, list):
        base = set(val1)
    else:
        base = {val1}

    if isinstance(val2, list):
        ext = set(val2)
    else:
        ext = {val2}
    return base.union(ext)


def apply_policy(metadata, policy):
    """
    Apply a metadata policy to a metadata statement

    :param metadata: A metadata statement
    :param policy: A metadata policy
    :return: A metadata statement that adheres to a metadata policy
    """

    metadata_set = set(metadata.keys())
    policy_set = set(policy.keys())

    # Metadata claims that there exists a policy for
    for claim in metadata_set.intersection(policy_set):
        if "subset_of" in policy[claim]:
            _val = set(policy[claim]['subset_of']).intersection(set(metadata[claim]))
            if _val:
                metadata[claim] = list(_val)
            else:
                raise PolicyError("{} not subset of {}".format(metadata[claim],
                                                               policy[claim]['subset_of']))
        elif "superset_of" in policy[claim]:
            if set(policy[claim]['superset_of']).difference(set(metadata[claim])):
                raise PolicyError("{} not superset of {}".format(metadata[claim],
                                                                 policy[claim]['superset_of']))
            else:
                pass
        elif "one_of" in policy[claim]:
            if isinstance(metadata[claim], list):
                _claim = None
                for c in metadata[claim]:
                    if c in policy[claim]['one_of']:
                        # Use the first that matches
                        _claim = c
                        break
                if _claim:
                    metadata[claim] = _claim
                else:
                    raise PolicyError(
                        "None of {} among {}".format(metadata[claim], policy[claim]['one_of']))
            else:
                if metadata[claim] in policy[claim]['one_of']:
                    pass
                else:
                    raise PolicyError(
                        "{} not among {}".format(metadata[claim], policy[claim]['one_of']))
        elif "add" in policy[claim]:
            metadata[claim] = list(union(metadata[claim], policy[claim]['add']))
        elif "value" in policy[claim]:
            metadata[claim] = policy[claim]["value"]

    # In policy but not in metadata
    for claim in policy_set.difference(metadata_set):
        if "default" in policy[claim]:
            metadata[claim] = policy[claim]['default']

        if "essential" in policy[claim] and policy[claim]["essential"]:
            raise PolicyError("Essential claim '{}' missing".format(claim))

        if "add" in policy[claim]:
            metadata[claim] = policy[claim]['add']

        if "value" in policy[claim]:
            metadata[claim] = policy[claim]['value']

    # All that are in metadata but not in policy should just remain

    return metadata


def diff2policy(new, old):
    res = {}
    for claim in set(new).intersection(set(old)):
        if new[claim] == old[claim]:
            continue
        else:
            res[claim] = {'value': new[claim]}

    for claim in set(new).difference(set(old)):
        if claim in ['contacts']:
            res[claim] = {'add': new[claim]}
        else:
            res[claim] = {'value': new[claim]}

    return res



import logging
from urllib.parse import urlencode
from urllib.parse import urlparse

from cryptojwt.jws.jws import factory
from oidcmsg.oidc import ProviderConfigurationResponse
from oidcservice.exception import ResponseError
from oidcservice.oidc.provider_info_discovery import ProviderInfoDiscovery

from fedservice.entity_statement.collect import branch2lists
from fedservice.entity_statement.collect import verify_self_signed_signature
from fedservice.entity_statement.construct import map_configuration_to_preference
from fedservice.entity_statement.utils import create_authority_hints
from fedservice.entity_statement.verify import eval_chain
from fedservice.exception import NoTrustedClaims

logger = logging.getLogger(__name__)


class FedProviderInfoDiscovery(ProviderInfoDiscovery):
    response_cls = ProviderConfigurationResponse
    request_body_type = 'jose'
    response_body_type = 'jose'

    def __init__(self, service_context, state_db, conf=None,
                 client_authn_factory=None, **kwargs):
        ProviderInfoDiscovery.__init__(
            self, service_context, state_db, conf=conf,
            client_authn_factory=client_authn_factory)

    def get_request_parameters(self, method="GET", **kwargs):
        try:
            _iss = kwargs["iss"]
        except KeyError:
            _iss = self.service_context.issuer

        qpart = {'iss': _iss}

        for param in ['sub', 'aud', 'prefetch']:
            try:
                qpart[param] = kwargs[param]
            except KeyError:
                pass

        p = urlparse(_iss)
        _qurl = '{}://{}/.well-known/openid-federation?{}'.format(
            p.scheme, p.netloc, urlencode(qpart))

        return {'url': _qurl, 'iss': _iss}

    def store_federation_info(self, statement, trust_root_id):
        """

        :param statement: A
            :py:class:`fedservice.entity_statement.statement.Statement` instance
        """
        # Only use trusted claims
        trusted_claims = statement.metadata
        if trusted_claims is None:
            raise NoTrustedClaims()
        _pi = self.response_cls(**trusted_claims)

        # Temporarily (?) taken out
        # if 'signed_jwks_uri' in _pi:
        #     _kb = KeyBundle(source=_pi['signed_jwks_uri'],
        #                     verify_keys=statement.signing_keys,
        #                     verify_ssl=False)
        #     _kb.do_remote()
        #     # Replace what was there before
        #     self.service_context.keyjar[self.service_context.issuer] = _kb

        self.service_context.provider_info = _pi
        self.service_context.federation_entity.federation = trust_root_id

    def update_service_context(self, statements, **kwargs):
        """
        The list of :py:class:`fedservice.entity_statement.statement.Statement` instances are
        stored in *provider_federations*.
        If the OP and RP only has one federation in common then the choice is
        easy and the name of the federation are stored in the *federation*
        attribute while the provider info are stored in the service_context.

        :param paths: A list of Statement instances
        :param kwargs: Extra Keyword arguments
        """

        _sc = self.service_context
        _fe = _sc.federation_entity

        possible = list(set([s.fo for s in statements]).intersection(_fe.tr_priority))

        _fe.provider_federations = possible
        _fe.op_statements = statements

        _fe.proposed_authority_hints = create_authority_hints(
            _fe.authority_hints, statements)

        if len(possible) == 1:
            for s in statements:
                if s.fo == possible[0]:
                    claims = s.metadata
                    _sc.provider_info = self.response_cls(**claims)
                    self._update_service_context(_sc.provider_info)
                    _sc.behaviour = map_configuration_to_preference(
                        _sc.provider_info, _sc.client_preferences)
        else:
            # Not optimal but a reasonable estimate for now
            claims = statements[0].metadata
            _pinfo = self.response_cls(**claims)
            _sc.behaviour = map_configuration_to_preference(
                _pinfo, _sc.client_preferences)

    def parse_response(self, info, sformat="", state="", **kwargs):
        # returns a list of Statement instances
        resp = self.parse_federation_response(info, state=state)

        if not resp:
            logger.error('Missing or faulty response')
            raise ResponseError("Missing or faulty response")

        return resp

    def parse_federation_response(self, response, **kwargs):
        """
        Takes a provider info response and parses it.
        If according to the info the OP has more then one federation
        in common with the client then the decision has to be handled higher up.
        For each Metadata statement that appears in the response, and was
        possible to parse, one
        :py:class:`fedservice.entity_statement.statement.Statement`
        instance is stored in the response by federation operator ID under the
        key 'fos'.

        :param response: A self-signed JWT containing an entity statement
        :returns: A list of lists of Statement instances. The innermost lists represents
        trust chains
        """

        _jwt = factory(response)
        entity_statement = _jwt.jwt.payload()
        entity_id = entity_statement['iss']

        _fe = self.service_context.federation_entity

        metadata = verify_self_signed_signature(response)
        _tree = _fe.collect_statement_chains(entity_id, metadata)
        _node = {entity_id: (response, _tree)}
        _chains = branch2lists(_node)
        for c in _chains:
            c.append(response)
        return [eval_chain(c, _fe.key_jar, 'openid_provider') for c in _chains]

    def get_response(self, *args, **kwargs):
        """

        :param args: Just ignore these
        :param kwargs:
        :return:
        """

        _fe = self.service_context.federation_entity
        self_signed_config = _fe.collector.get_configuration_information(kwargs["iss"])
        return self.parse_response(self_signed_config)



import json
import logging

from cryptojwt import key_bundle
from cryptojwt.jws.jws import factory
from cryptojwt.key_jar import KeyJar


logger = logging.getLogger(__name__)


class KeyBundle(key_bundle.KeyBundle):
    """
    Extended :py:class:`oidcmsg.key_bundle.KeyBundle` class that supports
    signed JWKS uris.
    """
    def __init__(self, keys=None, source="", cache_time=300, verify_ssl=True,
                 file_format="jwk", keytype="RSA", keyusage=None,
                 verify_keys=None):
        super(KeyBundle, self).__init__(keys=keys, source=source,
                                        cache_time=cache_time,
                                        verify_ssl=verify_ssl,
                                        fileformat=file_format,
                                        keytype=keytype, keyusage=keyusage)
        if verify_keys is not None:
            if isinstance(verify_keys, KeyJar):
                self.verify_keys = verify_keys
            else:
                self.verify_keys = KeyJar()
                self.verify_keys.import_jwks(verify_keys, '')

    def _parse_remote_response(self, response):
        """
        Parse simple JWKS or signed JWKS from the HTTP response.

        :param response: HTTP response from the 'jwks_uri' or 'signed_jwks_uri'
            endpoint
        :return: response parsed as JSON or None
        """
        # Check if the content type is the right one.
        try:
            if response.headers["Content-Type"] == 'application/json':
                logger.debug(
                    "Loaded JWKS: %s from %s" % (response.text, self.source))
                try:
                    return json.loads(response.text)
                except ValueError:
                    return None
            elif response.headers["Content-Type"] == 'application/jwt':
                logger.debug(
                    "Signed JWKS: %s from %s" % (response.text, self.source))
                _jws = factory(response.text)
                _resp = _jws.verify_compact(
                    response.text, keys=self.verify_keys.get_signing_key())
                return _resp
            else:
                logger.error('Wrong content type: {}'.format(
                    response.headers['Content-Type']))
                raise ValueError('Content-type mismatch')
        except KeyError:
            pass


import os

from oidcendpoint.util import importer

"""
        'federation_api': {
            'class': FSEntityStatementAPI,
            'kwargs': {
                'entity_id_pattern': BASE_URL + "/eid/{}",
                'client_authn_method': None,
                'base_path': 'base_data',
                'iss': "swamid.se",
                "url_prefix": BASE_URL + "/eid"
            }
        }
"""


class SigningService:
    def __init__(self, conf, cwd=''):
        self.issuer = {}
        self.wd = cwd
        self.web_cert_path = ""
        for attr, spec in conf.items():
            self.issuer[attr] = self.build_signing_service(spec)

    def build_signing_service(self, spec):
        spec['kwargs']['base_path'] = os.path.join(self.wd, spec['kwargs']["base_path"])
        if isinstance(spec["class"], str):
            _instance = importer(spec["class"])(**spec['kwargs'])
        else:
            _instance = spec["class"](**spec['kwargs'])

        return _instance


import json
import logging
import os
from urllib.parse import urlparse

from cryptojwt.key_jar import KeyJar

from fedservice.entity_statement.create import create_entity_statement

logger = logging.getLogger(__name__)


def mk_path(*args):
    _part = []
    for arg in args:
        if arg.startswith('https://') or arg.startswith('http://'):
            _ip = urlparse(arg)
            _part.append(format('_'.join(_ip.path[1:].split('/'))))
        else:
            _part.append(arg)
    _dir = "/".join(_part)

    if not os.path.isdir(_dir):
        return None
    else:
        return _dir


def read_info(dir, sub, typ='metadata'):
    file_name = os.path.join(dir, "{}.{}.json".format(sub, typ))
    if os.path.isfile(file_name):
        return json.loads(open(file_name).read())
    else:
        return None


def get_authority_hints(iss, sub, root_dir):
    _auth = read_info(os.path.join(root_dir, iss), sub, "authority")
    if _auth:
        return _auth
    else:
        return {}


def make_entity_statement(iss, root_dir='.', sub=""):
    kj = KeyJar()

    if iss.startswith('https://'):
        iss_id = iss
        iss = iss[len("https://"):]
    else:
        iss_id = "https://{}".format(iss)

    _jwks = read_info(os.path.join(root_dir, iss), iss, "jwks")
    kj.import_jwks(_jwks, iss_id)

    if not sub:
        sub = iss

    if sub.startswith('https://'):
        sub_id = sub
        sub = sub[len("https://"):]
    else:
        sub_id = "https://{}".format(sub)

    _jwks = read_info(os.path.join(root_dir, iss), sub, "jwks")
    kj.import_jwks(_jwks, sub_id)

    metadata = read_info(os.path.join(root_dir, iss), sub, "metadata")
    policy = read_info(os.path.join(root_dir, iss), sub, "policy")

    _auth = get_authority_hints(iss, sub, root_dir)
    if _auth:
        _jwt = create_entity_statement(iss_id, sub_id, kj, metadata, policy, _auth)
    else:
        _jwt = create_entity_statement(iss_id, sub_id, kj, metadata, policy)

    return _jwt



"""Support for TLS servers using self-signed certificates."""
import os


def add_ssc_support(federation_entity, **kwargs):
    """

    :param endpoint:
    :param kwargs:
    :return:
    """
    _collector = federation_entity.collector

    _collector.use_ssc = True
    _dir = kwargs.get("ssc_dir", "")
    if _dir.startswith("/"):
        _collector.ssc_dir = _dir
    else:
        _collector.ssc_dir = os.path.join(_collector.cwd, _dir)

    if not os.path.isdir(_collector.ssc_dir):
        os.makedirs(_collector.ssc_dir, exist_ok=True)

    _cert = kwargs.get("client_cert")
    _key = kwargs.get("client_key")

    if _cert and _key:
        _collector.httpc_params['cert'] = (_cert, _key)
    elif _cert:
        _collector.httpc_params['cert'] = _cert


import logging
import os
from urllib.parse import quote_plus
from urllib.parse import urlencode
from urllib.parse import urlparse

import requests
from cryptojwt import JWT
from cryptojwt import KeyJar
from cryptojwt.jwk import x5c_to_pems
from cryptojwt.jws.jws import factory
from cryptojwt.jwt import utc_time_sans_frac
from oidcmsg.exception import MissingPage
from requests.exceptions import SSLError

from fedservice.exception import UnknownCertificate
from .cache import ESCache

logger = logging.getLogger(__name__)


class FailedConfigurationRetrieval(Exception):
    pass


def construct_well_known_url(entity_id, typ):
    p = urlparse(entity_id)
    return '{}://{}/.well-known/{}'.format(p.scheme, p.netloc, typ)


def construct_tenant_well_known_url(entity_id, typ):
    p = urlparse(entity_id)
    return '{}://{}{}/.well-known/{}'.format(p.scheme, p.netloc, p.path, typ)


def unverified_entity_statement(signed_jwt):
    _jws = factory(signed_jwt)
    return _jws.jwt.payload()


def verify_self_signed_signature(config):
    """
    Verify signature. Will raise exception if signature verification fails.

    :param config: Signed JWT
    :return: Payload of the signed JWT
    """

    payload = unverified_entity_statement(config)
    keyjar = KeyJar()
    keyjar.import_jwks(payload['jwks'], payload['iss'])

    _jwt = JWT(key_jar=keyjar)
    _val = _jwt.unpack(config)
    return _val


def get_api_endpoint(config):
    return config['metadata']['federation_entity']["federation_api_endpoint"]


def construct_entity_statement_query(api_endpoint, issuer, subject):
    return "{}?{}".format(api_endpoint,
                          urlencode({
                              "iss": issuer,
                              "sub": subject
                          }))


def active(config):
    """
    Verifies that the signature of a configuration has not timed out.

    :param config:
    :return: True/False
    """
    return True


class Collector(object):
    def __init__(self, trust_anchors, http_cli=None, insecure=False,
                 allowed_delta=300, httpc_params=None, cwd=''):
        """

        :param trust_anchors:
        :param http_cli:
        :param allowed_delta:
        :param httpc_params: Additional parameters to pass to the HTTP client
            function
        """
        self.trusted_anchors = trust_anchors
        self.trusted_ids = set(trust_anchors.keys())
        self.config_cache = ESCache(300)
        self.entity_statement_cache = ESCache(300)
        self.http_cli = http_cli or requests.request
        self.allowed_delta = allowed_delta
        self.web_cert_path = None
        self.use_ssc = False
        self.ssc_dir = ""
        self.cwd = cwd

        self.httpc_params = httpc_params or {}
        if insecure:
            self.httpc_params["verify"] = False

    def get_entity_statement(self, api_endpoint, issuer, subject):
        """
        Get Entity Statement by one entity about another or about itself

        :param api_endpoint: The federation API endpoint
        :param issuer: Who should issue the entity statement
        :param subject: About whom the entity statement should be
        :return: A signed JWT
        """
        _url = construct_entity_statement_query(api_endpoint, issuer, subject)

        if self.use_ssc:
            signed_entity_statement = self.do_ssc_seq(_url, issuer)
        else:
            signed_entity_statement = self.get_signed_entity_statement(_url, self.httpc_params)

        return  signed_entity_statement

    def _cert_path(self, entity_id):
        return os.path.join(self.ssc_dir, "{}.pem".format(quote_plus(entity_id)))

    def store_ssc_cert(self, entity_statement, entity_id):
        """
        Convert a x5c value into a list of PEM formated certificates and write them to a file.

        :param entity_statement: An Entity statement that should contain a x5c parameter.
        :param entity_id: The ID of the subject of the the Entity Statement.
        :return: The path to the created file.
        """
        x5c = entity_statement.get("x5c")
        if x5c:
            _certs = x5c_to_pems(x5c)
            _cert_path = self._cert_path(entity_id)
            with open(_cert_path, "w") as fp:
                for c in _certs:
                    fp.write(c)

            return _cert_path
        return ""

    def get_cert_path(self, entity_id):
        """
        Construct a file path and verify that it points to a file.

        :param entity_id: Significant part of the file path.
        :return: The path to an existing file.
        """
        _cert_path = self._cert_path(entity_id)
        if os.path.isfile(_cert_path):
            return _cert_path

        return None

    def get_signed_entity_statement(self, url, httpc_args):
        """

        :param url: Target URL
        :param httpc_args: Arguments for the HTTP call.
        :return: Signed EntityStatement
        """

        response = self.http_cli("GET", url, **httpc_args)
        if response.status_code == 200:
            return response.text
        elif response.status_code == 404:
            raise MissingPage("No such page: '{}'".format(url))
        else:
            raise FailedConfigurationRetrieval()

    def do_ssc_seq(self, url, entity_id, self_signed=True):
        """
        Check that the self-signed certificate in the Entity Statement really is
        the one used by the entity.

        This involves 2 steps, since I'd rather find out if something amiss early then
        late.

        :param url: The page I want to access.
        :param entity_id: The ID for the entity I want the information on.
        :return:
        """
        logger.debug("Self-signed certification sequence")
        httpc_args = self.httpc_params.copy()

        # have I seen it before
        cert_path = self.get_cert_path(entity_id)

        _signed_entity_statement = ''
        if cert_path is None:
            logger.debug("No saved certificate")
            if self_signed:  # This only works for the self-signed entity statements
                # First get the Entity Statement without verifying the entity certificate
                httpc_args["verify"] = False

                _signed_entity_statement = self.get_signed_entity_statement(url, httpc_args)

                # The get the Entity Statement while using the certificate from the entity statement
                # to verify the HTTPS certificate
                cert_path = self.store_ssc_cert(unverified_entity_statement(_signed_entity_statement),
                                                 entity_id)
                if not cert_path:
                    logger.debug('No SSL certificate in the entity metadata')
            else:  # out of luck
                raise UnknownCertificate(entity_id)

        if cert_path:
            httpc_args["verify"] = cert_path
            _signed_entity_statement = self.get_signed_entity_statement(url, httpc_args)

        return _signed_entity_statement

    def get_configuration_information(self, entity_id):
        """
        Get configuration information about an entity from itself.
        The configuration information is in the format of an Entity Statement

        :param entity_id: About whom the entity statement should be
        :return: Configuration information as a signed JWT
        """

        _url = construct_well_known_url(entity_id, "openid-federation")
        logger.debug("Config URL: %s", _url)
        try:
            if self.use_ssc:
                self_signed_config = self.do_ssc_seq(_url, entity_id)
            else:
                self_signed_config = self.get_signed_entity_statement(_url, self.httpc_params)
        except MissingPage:  # if tenant involved
            _tenant_url = construct_tenant_well_known_url(entity_id, "openid-federation")
            logger.debug("Tenant config URL: %s", _tenant_url)
            if _tenant_url != _url:
                if self.use_ssc:
                    self_signed_config = self.do_ssc_seq(_tenant_url, entity_id)
                else:
                    self_signed_config = self.get_signed_entity_statement(_tenant_url,
                                                                          self.httpc_params)
            else:
                raise MissingPage("No such page: '{}'".format(_url))
        except SSLError as err:
            logger.error(err)
            raise

        return self_signed_config

    def get_federation_api_endpoint(self, intermediate):
        # In cache
        _info = self.config_cache[intermediate]
        if _info:
            fed_api_endpoint = get_api_endpoint(_info)
        else:
            fed_api_endpoint = None

        if not fed_api_endpoint:
            signed_entity_config = self.get_configuration_information(intermediate)
            if signed_entity_config is None:
                return None

            entity_config = verify_self_signed_signature(signed_entity_config)
            fed_api_endpoint = get_api_endpoint(entity_config)
            # update cache
            self.config_cache[intermediate] = entity_config

        return fed_api_endpoint

    def collect_intermediate(self, entity_id, intermediate, seen=None, max_superiors=10):
        """
        Collect information about an entity by another entity, the intermediate.
        This consist of first find the fed_api_endpoint URL for the intermediate and then
        asking the intermediate for its view of the entity.

        :param entity_id: The ID of the entity
        :param intermediate: The immediate superior
        :param seen: A list of intermediates that this process has seen. This to capture
            loops. Also used to control the allowed depth.
        :param max_superiors: The maximum number of superiors.
        :return:
        """
        # Should I stop when I reach the first trust anchor ?
        if entity_id == intermediate and entity_id in self.trusted_anchors:
            return None

        if seen is None:
            _seen = []
        else:
            _seen = seen[:]

        _seen.append(intermediate)
        # if len(_seen) > max_superiors:
        #     logger.warning("Reached max superiors. The path here was {}".format(_seen))
        #     return None

        # Try to get the entity statement from the cache
        cache_key = "{}!!{}".format(intermediate, entity_id)
        entity_statement = self.entity_statement_cache[cache_key]

        if entity_statement is not None:
            _now = utc_time_sans_frac()
            time_key = "{}!exp!{}".format(intermediate, entity_id)
            _exp = self.entity_statement_cache[time_key]
            if _now > (_exp - self.allowed_delta):
                logger.debug("Cached entity statement timed out")
                del self.entity_statement_cache[cache_key]
                del self.entity_statement_cache[time_key]
                entity_statement = None

        if entity_statement is None:
            fed_api_endpoint = self.get_federation_api_endpoint(intermediate)
            if fed_api_endpoint is None:
                raise SystemError('Could not find federation_api endpoint')
            logger.debug("Federation API endpoint: '{}' for '{}'".format(fed_api_endpoint,
                                                                        intermediate))
            entity_statement = self.get_entity_statement(fed_api_endpoint, intermediate,
                                                         entity_id)
            # entity_statement is a signed JWT
            statement = unverified_entity_statement(entity_statement)
            logger.debug("Unverified entity statement from {} about {}: {}".format(
                fed_api_endpoint, intermediate, statement))
            self.entity_statement_cache[cache_key] = entity_statement
            time_key = "{}!exp!{}".format(intermediate, entity_id)
            self.entity_statement_cache[time_key] = statement["exp"]

        if entity_statement:
            intermediate_statement = self.config_cache[intermediate]
            return entity_statement, self.collect_superiors(intermediate,
                                                            intermediate_statement,
                                                            seen=_seen,
                                                            max_superiors=max_superiors)
        else:
            return None

    def collect_superiors(self, entity_id, statement, seen=None, max_superiors=1, stop_at=""):
        """
        Collect superiors one level at the time

        :param entity_id: The entity ID
        :param statement: Metadata statement
        :param seen: A list of intermediates that this process has seen. This to capture
            loops. Also used to control the allowed depth.
        :param max_superiors: The maximum number of superiors.
        :param stop_at: The ID of the trust anchor at which the trust chain should stop.
        :return: Dictionary of superiors
        """
        superior = {}
        if seen is None:
            seen = []

        if 'authority_hints' not in statement:
            return superior
        elif statement['iss'] == stop_at:
            return superior

        for intermediate in statement['authority_hints']:
            if intermediate in seen:  # loop ?!
                logger.warning("Loop detected at {}".format(intermediate))
            logger.debug("Collect intermediate: %s", intermediate)
            superior[intermediate] = self.collect_intermediate(entity_id, intermediate, seen,
                                                               max_superiors)

        return superior


def branch2lists(node):
    res = []
    for issuer, branch in node.items():
        if branch is None:
            res.append([])
            continue

        (statement, node) = branch
        if not node:
            res = [[statement]]
            continue

        _lists = branch2lists(node)
        for l in _lists:
            l.append(statement)
        if not res:
            res = _lists
        else:
            res.extend(_lists)
    return res


def main(entity_id, anchors):
    collector = Collector(anchors)
    entity_config = collector.get_configuration_information(entity_id)
    _config = verify_self_signed_signature(entity_config)
    tree = entity_config, collector.collect_superiors(entity_id, _config)
    return tree


if __name__ == '__main__':
    leaf_id = "https://example.com/rp/fed"
    trusted_anchors = {
        "anchor_id": []  # Known public keys for a trusted anchor
    }

    tree = main(leaf_id, trusted_anchors)
    chains = branch2lists(tree)



import json
import os
from urllib.parse import unquote_plus

from fedservice.metadata_api import EntityStatementAPI


def read_info(dir, sub, typ='metadata'):
    file_name = os.path.join(dir, sub, "{}.json".format(typ))
    if os.path.isfile(file_name):
        return json.loads(open(file_name).read())
    else:
        return None


class FSEntityStatementAPI(EntityStatementAPI):
    def __init__(self, base_path, entity_id_pattern="https://{}", iss='', **kwargs):
        EntityStatementAPI.__init__(self, iss, entity_id_pattern)
        self.base_path = base_path
        if iss:
            # load own keys
            self.load_jwks(iss, iss, self.make_entity_id(iss))

        if 'url_prefix' in kwargs:
            self.url_prefix = kwargs['url_prefix']

    def load_jwks(self, sup, sub, sub_id):
        _jwks_file = os.path.join(self.base_path, sup, sub, "jwks.json")
        self.keyjar.import_jwks_as_json(open(_jwks_file).read(), sub_id)

    def gather_info(self, sub):
        iss_id = self.make_entity_id(self.iss)
        if iss_id not in self.keyjar:
            self.load_jwks(self.iss, self.iss, self.make_entity_id(sub))

        if sub.startswith("https%3A%2F%2F"):
            sub_id = unquote_plus(sub)
        else:
            sub_id = self.make_entity_id(sub)

        if sub_id not in self.keyjar:
            self.load_jwks(self.iss, sub, sub_id)

        data = {}
        for name, file in [("metadata", "metadata.json"),
                           ("metadata_policy", "policy.json"),
                           ("constraints", "constraints.json"),
                           ("authority_hints", "authority.json")]:
            metadata_file = os.path.join(self.base_path, self.iss, sub, file)
            if os.path.isfile(metadata_file):
                data[name] = json.loads(open(metadata_file).read())

        return data



from urllib.parse import unquote_plus
from urllib.parse import urlparse

from cryptojwt import KeyJar
from fedservice.entity_statement.create import create_entity_statement


class EntityStatementAPI:
    def __init__(self, iss, entity_id_pattern):
        self.iss = iss
        self.keyjar = KeyJar()
        self.entity_id_pattern = entity_id_pattern
        self.url_prefix = ''
        self.base_path = ""

    def gather_info(self, sub):
        raise NotImplementedError()

    def load_jwks(self, sup, sub, sub_id):
        raise NotImplementedError()

    def make_entity_id(self, netloc):
        return self.entity_id_pattern.format(netloc)

    def create_entity_statement(self, sub, **kwargs):
        _info = self.gather_info(sub)
        _info.update(kwargs)
        if sub.startswith("https"):
            return create_entity_statement(self.make_entity_id(self.iss), unquote_plus(sub),
                                           self.keyjar, **_info)

        return create_entity_statement(self.make_entity_id(self.iss), self.make_entity_id(sub),
                                       self.keyjar, **_info)





"""Configuration management for Signing Service"""

from typing import Dict

from oidcrp.logging import configure_logging
from oidcrp.util import get_http_params
from oidcrp.util import load_yaml_config
from oidcrp.util import lower_or_upper

try:
    from secrets import token_urlsafe as rnd_token
except ImportError:
    from oidcendpoint import rndstr as rnd_token


class Configuration:
    """Signing Service Configuration"""

    def __init__(self, conf: Dict) -> None:
        self.logger = configure_logging(config=conf.get('logging')).getChild(__name__)

        # server info
        self.domain = lower_or_upper(conf, "domain")
        self.port = lower_or_upper(conf, "port")

        # HTTP params
        _params = get_http_params(conf.get("http_params"))
        if _params:
            self.httpc_params = _params
        else:
            _params = {'verify', lower_or_upper(conf, "verify_ssl", True)}

        # web server config
        self.web_conf = lower_or_upper(conf, "webserver")

        srv_info = lower_or_upper(conf, "server_info", {})
        for entity, spec in srv_info.items():
            for key, arg in spec.items():
                if key == "kwargs":
                    _kw_args = {}
                    for attr, val in arg.items():
                        if attr in ["entity_id_pattern", "url_prefix"]:
                            _kw_args[attr] = val.format(domain=self.domain, port=self.port)
                        else:
                            _kw_args[attr] = val

                    spec["kwargs"] = _kw_args

        self.server_info = srv_info

    @classmethod
    def create_from_config_file(cls, filename: str):
        """Load configuration as YAML"""
        return cls(load_yaml_config(filename))



import base64
import json
import logging
from urllib.parse import quote_plus

import requests
from cryptojwt.jwt import JWT

logger = logging.getLogger(__name__)


def db_make_entity_statement(db_url, authn_info, key_jar, lifetime,
                             sign_alg='ES256', **kwargs):
    iss = kwargs['iss']

    try:
        sub = kwargs['sub']
    except KeyError:
        sub = iss

    # authn_method = ClientSecretBasic()
    # http_args = authn_method.construct({}, **authn_info)

    credentials = "{}:{}".format(authn_info['user'], authn_info['password'])
    authz = base64.b64encode(
        credentials.encode("utf-8")).decode("utf-8")
    http_args = {"headers": {"Authorization": "Basic {}".format(authz)}}

    if db_url.endswith('/'):
        _url = '{}{}'.format(db_url, quote_plus(sub))
    else:
        _url = '{}/{}'.format(db_url, quote_plus(sub))

    resp = requests.request('GET', _url, **http_args)

    if resp.status_code == 200:
        payload = json.loads(resp.text)

        packer = JWT(key_jar=key_jar, iss=iss, lifetime=lifetime,
                     sign=True, sign_alg=sign_alg)
        return packer.pack(payload)
    else:
        raise SystemError('DB not accessible "{}"'.format(resp.text))



import logging

from oidcendpoint.oidc import authorization
from oidcmsg import oidc
from oidcmsg.oidc import RegistrationRequest

logger = logging.getLogger(__name__)


class Authorization(authorization.Authorization):
    msg_type = oidc.AuthorizationRequest
    response_cls = oidc.AuthorizationResponse
    error_msg = oidc.ResponseMessage

    def __init__(self, endpoint_context, **kwargs):
        authorization.Authorization.__init__(self, endpoint_context, **kwargs)
        # self.pre_construct.append(self._pre_construct)
        # self.post_parse_request.append(self._post_parse_request)
        self.automatic_registration_endpoint = None

    def do_automatic_registration(self, entity_id):
        _fe = self.endpoint_context.federation_entity

        # get self-signed entity statement
        _sses = _fe.get_configuration_information(entity_id)

        # Collect all the trust chains, verify them and apply policies. return the result
        statements = _fe.collect_metadata_statements(_sses, "openid_relying_party")

        # pick one of the possible
        statement = _fe.pick_metadata(statements)

        # handle the registration request as in the non-federation case.
        req = RegistrationRequest(**statement.metadata)
        req['client_id'] = entity_id
        new_id = self.automatic_registration_endpoint.kwargs.get("new_id", False)
        response_info = self.automatic_registration_endpoint.process_request(req, new_id=new_id)
        try:
            return response_info["response_args"]["client_id"]
        except KeyError:
            return None

    def client_authentication(self, request, auth=None, **kwargs):

        _cid = request["client_id"]

        try:  # If this is a registered client then this should return some info
            self.endpoint_context.cdb[_cid]
        except KeyError:  # else go the federation way
            registered_client_id = self.do_automatic_registration(_cid)
            if registered_client_id is None:
                return {
                    'error': 'unauthorized_client',
                    'error_description': 'Unknown client'
                }
            else:
                request["client_id"] = registered_client_id
                kwargs["also_known_as"] = {_cid: registered_client_id}

        # then do client authentication
        return authorization.Authorization.client_authentication(
            self, request, auth, **kwargs)



import logging

from cryptojwt import as_unicode
from cryptojwt.jws.jws import factory
from cryptojwt.key_jar import KeyJar
from cryptojwt.key_jar import init_key_jar
from oidcendpoint.util import importer

from fedservice.entity_statement.collect import Collector
from fedservice.entity_statement.collect import branch2lists
from fedservice.entity_statement.create import create_entity_statement
from fedservice.entity_statement.policy import apply_policy
from fedservice.entity_statement.policy import combine_policy
from fedservice.entity_statement.policy import gather_policies
from fedservice.entity_statement.verify import eval_chain
from fedservice.entity_statement.verify import eval_policy_chain
from fedservice.utils import load_json

__author__ = 'Roland Hedberg'
__version__ = '0.6.0'

logger = logging.getLogger(__name__)


class FederationEntity(object):
    def __init__(self, entity_id, trusted_roots, authority_hints=None,
                 key_jar=None, default_lifetime=86400, httpd=None,
                 priority=None, entity_type='', opponent_entity_type='',
                 registration_type='', cwd='', httpc_params=None):
        self.collector = Collector(trust_anchors=trusted_roots, http_cli=httpd, cwd=cwd,
                                   httpc_params=httpc_params)
        self.entity_id = entity_id
        self.entity_type = entity_type
        self.opponent_entity_type = opponent_entity_type
        self.key_jar = key_jar or KeyJar(httpc_params=httpc_params)
        for iss, jwks in trusted_roots.items():
            self.key_jar.import_jwks(jwks, iss)
        self.authority_hints = authority_hints
        self.default_lifetime = default_lifetime
        self.tr_priority = priority or sorted(set(trusted_roots.keys()))
        self.registration_type = registration_type

    def collect_statement_chains(self, entity_id, statement):
        return self.collector.collect_superiors(entity_id, statement)

    def eval_chains(self, chains, entity_type='', apply_policies=True):
        """

        :param chains: A list of lists of signed JWT
        :param entity_type: The leafs entity type
        :param apply_policies: Apply metadata policies from the list on the the metadata of the
            leaf entity
        :return: List of Statement instances
        """
        if not entity_type:
            entity_type = self.opponent_entity_type

        return [eval_chain(c, self.key_jar, entity_type, apply_policies) for c in chains]

    def create_entity_statement(self, iss, sub, key_jar=None, metadata=None, metadata_policy=None,
                                authority_hints=None, lifetime=0, **kwargs):
        if not key_jar:
            key_jar = self.key_jar
        if not authority_hints:
            authority_hints = self.authority_hints
        if not lifetime:
            lifetime = self.default_lifetime

        return create_entity_statement(iss, sub, key_jar=key_jar, metadata=metadata,
                                       metadata_policy=metadata_policy,
                                       authority_hints=authority_hints, lifetime=lifetime, **kwargs)

    def get_configuration_information(self, subject_id):
        return self.collector.get_configuration_information(subject_id)

    def pick_metadata(self, statements):
        """
        Pick one statement out of the list of possible statements

        :param statements: A list of :py:class:`fedservice.entity_statement.statement.Statement
            instances
        :return: A :py:class:`fedservice.entity_statement.statement.Statement instance
        """
        if len(statements) == 1:
            # right now just pick the first:
            return statements[0]
        else:
            for fid in self.tr_priority:
                for statement in statements:
                    if statement.fo == fid:
                        return statement

        # Can only arrive here if the federations I got back and trust are not
        # in the priority list. So, just pick one

        return statements[0]

    def get_payload(self, self_signed_statement):
        _jws = as_unicode(self_signed_statement)
        _jwt = factory(_jws)
        return _jwt.jwt.payload()

    def collect_metadata_statements(self, self_signed_statement, metadata_type):
        """

        :param self_signed_statement: A Self signed Entity Statement
        :param metadata_type: One of the metadata types defined in the specification
        :return:
        """
        payload = self.get_payload(self_signed_statement)

        # collect trust chains
        _tree = self.collect_statement_chains(payload['iss'], payload)
        _node = {payload['iss']: (self_signed_statement, _tree)}
        _chains = branch2lists(_node)

        # verify the trust paths and apply policies
        return [eval_chain(c, self.key_jar, metadata_type) for c in _chains]


def create_federation_entity(entity_id, httpc_params=None, **kwargs):
    args = {"httpc_params": httpc_params}
    for param in ['trusted_roots', 'authority_hints']:
        try:
            args[param] = load_json(kwargs[param])
        except KeyError:
            pass

    if 'signing_keys' in kwargs:
        args['key_jar'] = init_key_jar(**kwargs['signing_keys'],
                                       owner=entity_id)
        args['key_jar'].httpc_params = httpc_params

    for param in ['entity_type', 'priority', 'opponent_entity_type',
                  'registration_type', 'cwd']:
        try:
            args[param] = kwargs[param]
        except KeyError:
            pass

    federation_entity = FederationEntity(entity_id, **args)

    add_ons = kwargs.get("add_on")
    for spec in add_ons.values():
        if isinstance(spec["function"], str):
            _func = importer(spec["function"])
        else:
            _func = spec["function"]
        _func(federation_entity, **spec["kwargs"])

    return federation_entity



""" Classes and functions used to describe information in an OpenID Connect Federation."""
import logging

from cryptojwt.exception import Expired
from cryptojwt.jwt import utc_time_sans_frac
from oidcmsg.exception import MissingRequiredAttribute
from oidcmsg.message import Message
from oidcmsg.message import OPTIONAL_LIST_OF_SP_SEP_STRINGS
from oidcmsg.message import OPTIONAL_LIST_OF_STRINGS
from oidcmsg.message import REQUIRED_LIST_OF_STRINGS
from oidcmsg.message import SINGLE_OPTIONAL_ANY
from oidcmsg.message import SINGLE_OPTIONAL_INT
from oidcmsg.message import SINGLE_OPTIONAL_JSON
from oidcmsg.message import SINGLE_OPTIONAL_STRING
from oidcmsg.message import SINGLE_REQUIRED_INT
from oidcmsg.message import SINGLE_REQUIRED_STRING
from oidcmsg.message import msg_ser
from oidcmsg.oidc import JsonWebToken
from oidcmsg.oidc import ProviderConfigurationResponse
from oidcmsg.oidc import RegistrationResponse
from oidcmsg.oidc import SINGLE_OPTIONAL_BOOLEAN
from oidcmsg.oidc import SINGLE_OPTIONAL_DICT
from oidcmsg.oidc import deserialize_from_one_of
from oidcmsg.oidc import dict_deser
from oidcmsg.oidc import msg_ser_json

from fedservice.exception import WrongSubject

SINGLE_REQUIRED_DICT = (dict, True, msg_ser_json, dict_deser, False)

LOGGER = logging.getLogger(__name__)


def registration_response_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into a RegistrationResponse."""
    return deserialize_from_one_of(val, RegistrationResponse, sformat)


def provider_info_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into a ProviderConfigurationResponse."""
    return deserialize_from_one_of(val, ProviderConfigurationResponse, sformat)


OPTIONAL_CLIENT_METADATA = (Message, False, msg_ser, registration_response_deser, False)
OPTIONAL_PROVIDER_METADATA = (Message, False, msg_ser, provider_info_deser, False)


class AuthorizationServerMetadata(Message):
    """Metadata for an OAuth2 Authorization Server."""
    c_param = {
        "issuer": SINGLE_REQUIRED_STRING,
        "authorization_endpoint": SINGLE_OPTIONAL_STRING,
        "token_endpoint": SINGLE_OPTIONAL_STRING,
        "jwks_uri": SINGLE_OPTIONAL_STRING,
        "registration_endpoint": SINGLE_OPTIONAL_STRING,
        "scopes_supported": SINGLE_OPTIONAL_JSON,
        "response_types_supported": SINGLE_OPTIONAL_JSON,
        "response_modes_supported": SINGLE_OPTIONAL_JSON,
        "grant_types_supported": SINGLE_OPTIONAL_JSON,
        "token_endpoint_auth_methods_supported": SINGLE_OPTIONAL_JSON,
        "token_endpoint_auth_signing_alg_values_supported": SINGLE_OPTIONAL_JSON,
        "service_documentation": SINGLE_OPTIONAL_STRING,
        "ui_locales_supported": SINGLE_OPTIONAL_JSON,
        "op_policy_uri": SINGLE_OPTIONAL_STRING,
        "revocation_endpoint": SINGLE_OPTIONAL_STRING,
        "revocation_endpoint_auth_methods_supported": SINGLE_OPTIONAL_JSON,
        "revocation_endpoint_auth_signing_alg_values_supported": SINGLE_OPTIONAL_JSON,
        "introspection_endpoint": SINGLE_OPTIONAL_STRING,
        "introspection_endpoint_auth_methods_supported": SINGLE_OPTIONAL_JSON,
        "introspection_endpoint_auth_signing_alg_values_supported": SINGLE_OPTIONAL_JSON,
        "code_challenge_methods_supported": SINGLE_OPTIONAL_JSON
    }


def auth_server_info_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into an AuthorizationServerMetadata."""
    return deserialize_from_one_of(val, AuthorizationServerMetadata, sformat)


OPTIONAL_AUTH_SERVER_METADATA = (Message, False, msg_ser, auth_server_info_deser, False)


class NamingConstraints(Message):
    """Class representing naming constraints."""
    c_param = {
        "permitted": OPTIONAL_LIST_OF_STRINGS,
        "excluded": OPTIONAL_LIST_OF_STRINGS
    }


def naming_constraints_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into an NamingConstraints."""
    return deserialize_from_one_of(val, NamingConstraints, sformat)


SINGLE_OPTIONAL_NAMING_CONSTRAINTS = (Message, False, msg_ser, naming_constraints_deser, False)


class FederationEntity(Message):
    """Class representing a Federation Entity."""
    c_param = {
        "federation_api_endpoint": SINGLE_OPTIONAL_STRING,
        "name": SINGLE_OPTIONAL_STRING,
        "contacts": OPTIONAL_LIST_OF_STRINGS,
        "policy_url": SINGLE_OPTIONAL_STRING,
        "homepage_uri": SINGLE_OPTIONAL_STRING,
        "trust_anchor_id": SINGLE_OPTIONAL_STRING
    }


def federation_entity_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into a FederationEntity."""
    return deserialize_from_one_of(val, FederationEntity, sformat)


OPTIONAL_FEDERATION_ENTITY_METADATA = (Message, False, msg_ser,
                                       federation_entity_deser, False)


class OauthClientMetadata(Message):
    """Metadata for an OAuth2 Client."""
    c_param = {
        "redirect_uris": OPTIONAL_LIST_OF_STRINGS,
        "token_endpoint_auth_method": SINGLE_OPTIONAL_STRING,
        "grant_type": OPTIONAL_LIST_OF_STRINGS,
        "response_types": OPTIONAL_LIST_OF_STRINGS,
        "client_name": SINGLE_OPTIONAL_STRING,
        "client_uri": SINGLE_OPTIONAL_STRING,
        "logo_uri": SINGLE_OPTIONAL_STRING,
        "scope": OPTIONAL_LIST_OF_SP_SEP_STRINGS,
        "contacts": OPTIONAL_LIST_OF_STRINGS,
        "tos_uri": SINGLE_OPTIONAL_STRING,
        "policy_uri": SINGLE_OPTIONAL_STRING,
        "jwks_uri": SINGLE_OPTIONAL_STRING,
        "jwks": SINGLE_OPTIONAL_JSON,
        "software_id": SINGLE_OPTIONAL_STRING,
        "software_version": SINGLE_OPTIONAL_STRING
    }


def oauth_client_metadata_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into a OauthClientMetadata."""
    return deserialize_from_one_of(val, OauthClientMetadata, sformat)


OPTIONAL_OAUTH_CLIENT_METADATA = (Message, False, msg_ser,
                                  oauth_client_metadata_deser, False)


class OauthClientInformationResponse(OauthClientMetadata):
    """The information returned by a OAuth2 Server about an OAuth2 client."""
    c_param = OauthClientMetadata.c_param.copy()
    c_param.update({
        "client_id": SINGLE_REQUIRED_STRING,
        "client_secret": SINGLE_OPTIONAL_STRING,
        "client_id_issued_at": SINGLE_OPTIONAL_INT,
        "client_secret_expires_at": SINGLE_OPTIONAL_INT
    })

    def verify(self, **kwargs):
        super(OauthClientInformationResponse, self).verify(**kwargs)

        if "client_secret" in self:
            if "client_secret_expires_at" not in self:
                raise MissingRequiredAttribute(
                    "client_secret_expires_at is a MUST if client_secret is present")


class Metadata(Message):
    """The different types of metadata that an entity in a federation can belong to."""
    c_param = {
        'openid_relying_party': OPTIONAL_CLIENT_METADATA,
        'openid_provider': OPTIONAL_PROVIDER_METADATA,
        "oauth_authorization_server": OPTIONAL_AUTH_SERVER_METADATA,
        "oauth_client": OPTIONAL_OAUTH_CLIENT_METADATA,
        "federation_entity": OPTIONAL_FEDERATION_ENTITY_METADATA
    }


def metadata_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into a MetadataPolicy."""
    return deserialize_from_one_of(val, Metadata, sformat)

SINGLE_REQUIRED_METADATA = (Message, True, msg_ser, metadata_deser, False)
SINGLE_OPTIONAL_METADATA = (Message, False, msg_ser, metadata_deser, False)


class MetadataPolicy(Message):
    """The metadata policy verbs."""
    c_param = {
        "subset_of": OPTIONAL_LIST_OF_STRINGS,
        "one_of": OPTIONAL_LIST_OF_STRINGS,
        "superset_of": OPTIONAL_LIST_OF_STRINGS,
        "add": OPTIONAL_LIST_OF_STRINGS,
        "value": SINGLE_OPTIONAL_ANY,
        "default": SINGLE_OPTIONAL_ANY,
        "essential": SINGLE_OPTIONAL_BOOLEAN
    }


def metadata_policy_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into a MetadataPolicy."""
    return deserialize_from_one_of(val, MetadataPolicy, sformat)


SINGLE_REQUIRED_METADATA_POLICY = (Message, True, msg_ser, metadata_policy_deser, False)
SINGLE_OPTIONAL_METADATA_POLICY = (Message, False, msg_ser, metadata_policy_deser, False)


class Constraints(Message):
    """The types of constraints that can be applied to a trust chain."""
    c_param = {
        "max_path_length": SINGLE_OPTIONAL_INT,
        "naming_constraints": SINGLE_OPTIONAL_NAMING_CONSTRAINTS
    }


def constrains_deser(val, sformat="json"):
    """Deserializes a JSON object (most likely) into a Constraints."""
    return deserialize_from_one_of(val, Constraints, sformat)


SINGLE_REQUIRED_CONSTRAINS = (Message, True, msg_ser, constrains_deser, False)
SINGLE_OPTIONAL_CONSTRAINS = (Message, False, msg_ser, constrains_deser, False)


class EntityStatement(JsonWebToken):
    """The Entity Statement"""
    c_param = JsonWebToken.c_param.copy()
    c_param.update({
        "sub": SINGLE_REQUIRED_STRING,
        'iss': SINGLE_REQUIRED_STRING,
        'exp': SINGLE_REQUIRED_INT,
        'iat': SINGLE_REQUIRED_INT,
        "jti": SINGLE_OPTIONAL_STRING,
        'authority_hints': OPTIONAL_LIST_OF_STRINGS,
        'metadata': SINGLE_OPTIONAL_METADATA,
        'metadata_policy': SINGLE_OPTIONAL_METADATA_POLICY,
        'constraints': SINGLE_OPTIONAL_CONSTRAINS,
        'sub_meta': SINGLE_OPTIONAL_DICT,
        'jwks': SINGLE_OPTIONAL_DICT,
        "crit": OPTIONAL_LIST_OF_STRINGS,
        "policy_language_crit": OPTIONAL_LIST_OF_STRINGS
    })


class TrustMark(JsonWebToken):
    c_param = JsonWebToken.c_param.copy()
    c_param.update({
        "sub": SINGLE_REQUIRED_STRING,
        'iss': SINGLE_REQUIRED_STRING,
        'iat': SINGLE_REQUIRED_INT,
        "id": SINGLE_REQUIRED_STRING,
        "mark": SINGLE_OPTIONAL_STRING,
        "exp": SINGLE_OPTIONAL_INT,
        "ref": SINGLE_OPTIONAL_STRING
    })

    def verify(self, **kwargs):
        super(TrustMark, self).verify(**kwargs)

        entity_id = kwargs.get("entity_id")

        if entity_id is not None and entity_id != self["sub"]:
            raise WrongSubject("Mismatch between subject in trust mark and entity_id of entity")

        exp = kwargs.get("exp", 0)
        if exp:
            _now = utc_time_sans_frac()
            if _now > exp:  # have passed the time of expiration
                raise Expired()

        return True



def create_authority_hints(default_hints, statements):
    """

    :param default_hints: The authority hints provided to the entity at startup
    :param statements: A list of Statement instances
    :return: An authority_hints dictionary
    """

    intermediates = {statement.verified_chain[-1]['iss'] for statement in statements}
    return list(set(default_hints).intersection(intermediates))



import json
import logging
import ssl
import sys

from oidcservice.exception import ResponseError

logger = logging.getLogger(__name__)


def load_json(file_name):
    with open(file_name) as fp:
        js = json.load(fp)
    return js


def fed_parse_response(instance, info, sformat="", state="", **kwargs):
    if sformat in ['jose', 'jws', 'jwe']:
        resp = instance.post_parse_response(info, state=state)

        if not resp:
            logger.error('Missing or faulty response')
            raise ResponseError("Missing or faulty response")

        return resp

    else:
        return instance.parse_response(info, sformat, state, **kwargs)



import logging

from oidcmsg.oidc import RegistrationRequest
from oidcservice.oidc.provider_info_discovery import PROVIDER2PREFERENCE

from fedservice.entity_statement.policy import apply_policy


logger = logging.getLogger(__name__)


def translate_configuration(conf):
    """
    Map a Provider Configuration response into a metadata policy

    :param conf: Attribute,value pairs from a Provider Configuration response
    :return: Attribute,value pairs useful when constructing a Client
        registration requests
    """
    policy = {}
    cls = RegistrationRequest
    for pro, pref in PROVIDER2PREFERENCE.items():
        try:
            _allow = conf[pro]
        except KeyError:
            pass
        else:
            if pref == "scope":
                policy[pref] = {'subset_of': _allow}
            elif isinstance(cls.c_param[pref][0], list):
                policy[pref] = {'subset_of': _allow}
            else:
                policy[pref] = {'one_of': _allow}

    return policy


def list_to_singleton(args, cls):
    """
    For attributes that are listed as singletons but have values that are
    lists. Replace the list with the first element in the list.

    :param args: The attribute,value pairs as a dictionary
    :param cls: A class instance that can be used to get information about
        the value types for specific attributes.
    :return: A modified attribute,values dictionary
    """
    ci = cls()
    res = {}
    for key, val in args.items():
        if isinstance(val, list):
            try:
                _typ = ci.c_param[key][0]
            except KeyError:
                res[key] = val
            else:
                if isinstance(_typ, list):
                    res[key] = val
                else:
                    res[key] = val[0]
        else:
            res[key] = val
    return res


def map_configuration_to_preference(provider_configuration, client_preference):
    """
    Given a provider configuration and client registration attribute
    preferences initiate a ClientRegistration class.

    :param provider_configuration: A ProviderConfigurationResponse
    :param client_preference: An dictionary with client preferences
    :return: A ClientRegistration instance
    """
    _allowed = translate_configuration(provider_configuration)
    args = apply_policy(client_preference, _allowed)

    return list_to_singleton(args, RegistrationRequest)



from oidcendpoint import endpoint_context


class EndpointContext(endpoint_context.EndpointContext):
    def __init__(self, conf, keyjar=None, client_db=None, session_db=None,
                 cwd='', cookie_dealer=None, federation_entity=None):
        endpoint_context.EndpointContext.__init__(
            self, conf, keyjar=keyjar, client_db=client_db,
            session_db=session_db, cwd=cwd, cookie_dealer=cookie_dealer)

        self.federation_entity = federation_entity



import logging

from cryptojwt.jws.jws import factory

from fedservice.entity_statement.collect import unverified_entity_statement
from fedservice.entity_statement.policy import combine_policy

from fedservice.entity_statement.policy import apply_policy
from oidcmsg.oidc import RegistrationRequest
from oidcmsg.oidc import RegistrationResponse
from oidcservice.exception import ResponseError
from oidcservice.oidc.registration import Registration

from fedservice.entity_statement.collect import branch2lists
from fedservice.entity_statement.verify import eval_policy_chain

logger = logging.getLogger(__name__)


class FedRegistration(Registration):
    msg_type = RegistrationRequest
    response_cls = RegistrationResponse
    endpoint_name = 'registration_endpoint'
    endpoint = 'registration'
    request_body_type = 'jose'
    response_body_type = 'jose'

    def __init__(self, service_context, state_db, conf=None,
                 client_authn_factory=None, **kwargs):
        Registration.__init__(self, service_context, state_db, conf=conf,
                              client_authn_factory=client_authn_factory)
        #
        self.post_construct.append(self.create_entity_statement)

    @staticmethod
    def carry_receiver(request, **kwargs):
        if 'receiver' in kwargs:
            return request, {'receiver': kwargs['receiver']}
        else:
            return request, {}

    def create_entity_statement(self, request_args, service=None, **kwargs):
        """
        Create a self signed entity statement

        :param request_args:
        :param service:
        :param kwargs:
        :return:
        """

        _fe = self.service_context.federation_entity
        _md = {_fe.entity_type: request_args.to_dict()}
        return _fe.create_entity_statement(
            iss=_fe.entity_id, sub=_fe.entity_id, metadata=_md, key_jar=_fe.key_jar,
            authority_hints=_fe.proposed_authority_hints)

    def parse_response(self, info, sformat="", state="", **kwargs):
        resp = self.parse_federation_registration_response(info, **kwargs)

        if not resp:
            logger.error('Missing or faulty response')
            raise ResponseError("Missing or faulty response")

        return resp

    def parse_federation_registration_response(self, resp, **kwargs):
        """
        Receives a dynamic client registration response,

        :param resp: An entity statement instance
        :return: A set of metadata claims
        """
        _sc = self.service_context
        _fe = _sc.federation_entity

        # Can not collect trust chain. Have to verify the signed JWT with keys I have

        kj = self.service_context.federation_entity.key_jar
        _jwt = factory(resp)
        entity_statement = _jwt.verify_compact(resp, keys=kj.get_jwt_verify_keys(_jwt.jwt))

        _fo = entity_statement['metadata']["federation_entity"]["trust_anchor_id"]
        chosen = None
        for op_statement in _fe.op_statements:
            if op_statement.fo == _fo:
                chosen = op_statement
                break

        if not chosen:
            raise ValueError('No matching federation operator')

        # based on the Federation ID, conclude which OP config to use
        op_claims = chosen.metadata
        # _sc.trust_path = (chosen.fo, _fe.op_paths[statement.fo][0])
        _sc.provider_info = self.response_cls(**op_claims)

        # To create RPs metadata collect the trust chains
        tree = {}
        for ah in _fe.authority_hints:
            tree[ah] = _fe.collector.collect_intermediate(_fe.entity_id, ah)

        _node = {_fe.entity_id: (resp, tree)}
        chains = branch2lists(_node)

        # Get the policies
        policy_chains_tup = [eval_policy_chain(c, _fe.key_jar, _fe.entity_type) for c in chains]
        _policy = combine_policy(policy_chains_tup[0][1],
                                 entity_statement['metadata_policy'][_fe.entity_type])
        print("Combined policy: {}".format(_policy))
        _query = unverified_entity_statement(kwargs["request_body"])["metadata"][_fe.entity_type]
        _sc.registration_response = apply_policy(_query, _policy)
        return _sc.registration_response

    def update_service_context(self, resp, **kwargs):
        Registration.update_service_context(self, resp, **kwargs)
        _fe = self.service_context.federation_entity
        _fe.iss = resp['client_id']

    def get_response_ext(self, url, method="GET", body=None, response_body_type="",
                     headers=None, **kwargs):
        """

        :param url:
        :param method:
        :param body:
        :param response_body_type:
        :param headers:
        :param kwargs:
        :return:
        """

        _collector = self.service_context.federation_entity.collector

        httpc_args = _collector.httpc_parms.copy()
        # have I seen it before
        cert_path = _collector.get_cert_path(self.service_context.provider_info["issuer"])
        if cert_path:
            httpc_args["verify"] = cert_path

        try:
            resp = _collector.http_cli(method, url, data=body, headers=headers, **httpc_args)
        except Exception as err:
            logger.error('Exception on request: {}'.format(err))
            raise

        if 300 <= resp.status_code < 400:
            return {'http_response': resp}

        if "keyjar" not in kwargs:
            kwargs["keyjar"] = self.service_context.keyjar
        if not response_body_type:
            response_body_type = self.response_body_type

        if response_body_type == 'html':
            return resp.text

        if body:
            kwargs['request_body'] = body

        return self.parse_response(resp, response_body_type, **kwargs)



import logging

from fedservice.message import EntityStatement
from oidcendpoint.endpoint import Endpoint
from oidcendpoint.oidc import provider_config
from oidcmsg import oidc

logger = logging.getLogger(__name__)


class FederationAPI(Endpoint):
    request_cls = oidc.Message
    response_cls = EntityStatement
    response_format = 'jws'

    def __init__(self, endpoint_context, **kwargs):
        Endpoint.__init__(self, endpoint_context, **kwargs)
        self.post_construct.append(self.create_entity_statement)
        self.metadata_api = None

    def process_request(self, request=None, **kwargs):

        return {'response_args': self.endpoint_context.provider_info.copy()}

    def create_entity_statement(self, request_args, request=None, **kwargs):
        """
        Create a self signed entity statement

        :param request_args:
        :param request:
        :param kwargs:
        :return:
        """

        _fe = self.endpoint_context.federation_entity
        _md = {_fe.entity_type: request_args.to_dict()}
        return _fe.create_entity_statement(_fe.entity_id, sub=_fe.entity_id, metadata=_md)
